@@ -1,9 +1,9 @@
-'''Utilities for processing a continuous stream of data.
+'''Process a data stream through I/O-bound operations with concurrency.
 
 An input data stream goes through a series of operations.
 The target use case is that one or more operations is I/O bound,
-hence can benefit from multi-thread concurrency.
-These operations are triggered via `transform`.
+hence can benefit from concurrency via threading or asyncio.
+These operations are called `transform`s.
 
 The other operations are typically light weight and supportive
 of the main (concurrent) operation. These operations perform batching,
@@ -13,52 +13,70 @@ unbatching, buffering, filtering, logging, etc.
 Basic usage
 ===========
 
-In a typical use case, one starts with a `Stream` object, and calls
-its methods in a "chained" fashion:
+In a typical use case, one starts with a `Streamer` object, places it under
+context management, and calls its methods in a "chained" fashion:
 
     data = range(100)
-    pipeline = (
-        Stream(data)
-        .batch(10)
-        .transform(my_op_that_takes_a_batch, workers=4)
-        .unbatch()
-        )
+    with Streamer(data) as stream:
+        pipeline = (
+            stream
+            .batch(10)
+            .transform(my_op_that_takes_a_batch, concurrency=4)
+            .unbatch()
+            )
 
-After this setup, there are several ways to use the object `pipeline`.
+The methods `batch`, `transform`, and `unbatch` all modify the object `stream`
+"in-place" and return the original object, hence it's fine to add these operations
+one at a time, and it's not necessary to the intermediate results to new identifiers.
+The above is equivalent to the following:
 
-    1. Since `pipeline` is an Iterable and an Iterator, we can use it as such.
+    with Streamer(range(100)) as stream:
+        stream.batch(10)
+        stream.transform(my_op_that_takes_a_batch, concurrency=4)
+        stream.unbatch()
+        pipeline = stream
+
+After this setup, there are several ways to use the object `stream`.
+
+    1. Since `stream` is an Iterable and an Iterator, we can use it as such.
        Most naturally, iterate over it and process each element however
        we like.
 
-       We can of couse also provide `pipeline` as a parameter where an iterable
+       We can of couse also provide `stream` as a parameter where an iterable
        or iterator is expected. For example, the `mpservice.mpserver.Server`
        class has a method `stream` that expects an iterable, hence
        we can do things like
 
             server = Server(...)
             with server:
-                pipeline = ...
-                pipeline = server.stream(pipeline)
-                pipeline = pipeline.transform(yet_another_io_op)
+                for y in server.stream(stream):
+                    ...
+       Note that `server.stream(...)` does not produce a `Streamer` object.
+       If we want to put it in subsequent operations, simply turn it into a
+       `Streamer` object:
+
+                pipeline = Streamer(server.stream(stream))
+                pipeline.transform(yet_another_io_op)
+                ...
 
     2. If the stream is not too long (not "big data"), we can convert it to
        a list by the method `collect`:
 
-            result = pipeline.collect()
+            result = stream.collect()
 
-    3. If we don't need the elements coming out of `pipeline`, but rather
+    3. If we don't need the elements coming out of `stream`, but rather
        just need the original data (`data`) to flow through all the operations
        of the pipeline (e.g. if the last "substantial" operation is inserting
-       the data into a database), we can "drain" the pipeline:
+       the data into a database), we can "drain" the stream:
 
-            n = pipeline.drain()
+            n = stream.drain()
 
        where the returned `n` is the number of elements coming out of the
        last operation in the pipeline.
 
     4. We can continue to add more operations to the pipeline, for example,
 
-            pipeline = pipeline.transform(another_op, workers=3)
+            stream.transform(another_op, concurrency=3)
 
 ======================
 Handling of exceptions
@@ -124,8 +142,10 @@ multiple results for a single input element (if it produces a list, say,
 that list would be the result for the single input.)
 If the operation is mainly for the side effect, e.g.
 saving data in files or a database, hence there isn't much useful result,
-then the result could be `None`, which is not a problem. Regardless,
+then the result could be `None`, which is perfectly valid. Regardless,
 the returned `None`s will still become the resultant stream.
+
+Reference (for an early version of the code): https://zpz.github.io/blog/stream-processing/
 '''
 
 # Iterable vs iterator
@@ -156,27 +176,30 @@ from __future__ import annotations
 # https://stackoverflow.com/a/49872353
 # Will no longer be needed in Python 3.10.
 
-import collections.abc
+import asyncio
 import concurrent.futures
-import functools
 import logging
 import random
 import threading
+import traceback
+from queue import Queue, Empty as QueueEmpty, Full as QueueFull
+from time import sleep
 from typing import (
     Callable, TypeVar, Union, Optional,
     Iterable, Iterator,
-    Tuple, Type,
+    Tuple, Awaitable,
 )
 
-from overrides import EnforceOverrides, overrides
+from overrides import EnforceOverrides, overrides, final
 
-from .util import is_exception, IterQueue, EventUpstreamer, MAX_THREADS
+from .remote_exception import RemoteException, exit_err_msg
+from .util import is_exception, is_async, MAX_THREADS
 
 
 logger = logging.getLogger(__name__)
 
-T = TypeVar('T')
-TT = TypeVar('TT')
+T = TypeVar('T')      # indicates input data element
+TT = TypeVar('TT')    # indicates output after an op on `T`
 
 
 def _default_peek_func(i, x):
@@ -185,8 +208,75 @@ def _default_peek_func(i, x):
     print(x)
 
 
-class StreamMixin(EnforceOverrides):
+class Streamer(EnforceOverrides):
+    def __init__(self, instream: Union[Iterator, Iterable], /):
+        self.streamlets = [Stream(instream)]
+
+    def __enter__(self):
+        if self.streamlets[0]._thread_pool is None:
+            self.streamlets[0]._thread_pool = concurrent.futures.ThreadPoolExecutor()
+        self.streamlets[0]._started.set()
+        return self
+
+    def __exit__(self, exc_type=None, exc_value=None, exc_tb=None):
+        self.streamlets[0]._stopped.set()
+        self.streamlets[0]._started.clear()
+        msg = exit_err_msg(self, exc_type, exc_value, exc_tb)
+        if msg:
+            logger.error(msg)
+        self.streamlets[0]._thread_pool.shutdown()
+
+    @final
+    def __iter__(self):
+        if not self.streamlets[0]._started.is_set():
+            raise RuntimeError("iteration must be started within context manager, i.e. within a 'with ...:' block")
+        return self
+
+    @final
+    def __next__(self):
+        return self.streamlets[-1].__next__()
+
+    def collect(self) -> list:
+        '''
+        Gather the stream in a list. DO NOT do this on big data.
+        '''
+        return list(self)
+
+    def drain(self) -> Tuple[int, int]:
+        '''Drain off the stream.
+
+        Return a tuple of the number of elements processed
+        as well as the number of exceptions (hopefully 0!).
+
+        The number of exceptions could be non-zero only if
+        upstream transformers have set `return_exceptions` to True.
+        Otherwise, any exception would have propagated and
+        prevented this method from completing.
+        '''
+        n = 0
+        nexc = 0
+        for v in self:
+            n += 1
+            if is_exception(v):
+                nexc += 1
+        return n, nexc
+
+    def drop_if(self, func: Callable[[int, T], bool]):
+        '''
+        `func`: a function that takes the data element index (`self.index`)
+            along with the element value, and returns `True` if the element
+            should be skipped, that is, not included in the output stream.
+        '''
+        self.streamlets.append(Dropper(self.streamlets[-1], func))
+        return self
+
     def drop_exceptions(self):
+        '''
+        Used to skip exception objects that are produced in an upstream
+        transform that has `return_exceptions=True`. This way,
+        the previous op allows exceptions (i.e. do not crash), and
+        this op removes the exception objects from the output stream.
+        '''
         return self.drop_if(lambda i, x: is_exception(x))
 
     def drop_nones(self):
@@ -197,6 +287,10 @@ class StreamMixin(EnforceOverrides):
         return self.drop_if(lambda i, x: i < n)
 
     def keep_if(self, func: Callable[[int, T], bool]):
+        '''Keep an element in the stream only if a condition is met.
+
+        This is the opposite of `drop_if`.
+        '''
         return self.drop_if(lambda i, x: not func(i, x))
 
     def keep_every_nth(self, nth: int):
@@ -208,6 +302,33 @@ class StreamMixin(EnforceOverrides):
         rand = random.random
         return self.keep_if(lambda i, x: rand() < frac)
 
+    def head(self, n: int):
+        '''
+        Takes the first `n` elements and ignore the rest.
+        '''
+        # This does not delegate to `keep`, because `keep`
+        # would need to walk throught the entire stream,
+        # which is not needed for `head`.
+        self.streamlets.append(Header(self.streamlets[-1], n))
+        return self
+
+    def peek(self, func: Callable[[int, T], None] = None):
+        '''Take a peek at the data element *before* it is sent
+        on for processing.
+
+        The function `func` takes the data index (`self.index`)
+        and the data element. Typical actions include print out
+        info or save the data for later inspection. Usually this
+        function should not modify the data element in the stream.
+
+        User has flexibilities in `func`, e.g. to not print anything
+        under certain conditions.
+        '''
+        if func is None:
+            func = _default_peek_func
+        self.streamlets.append(Peeker(self.streamlets[-1], func))
+        return self
+
     def peek_every_nth(self, nth: int, peek_func: Callable[[int, T], None] = None):
         assert nth > 0
 
@@ -242,110 +363,56 @@ class StreamMixin(EnforceOverrides):
 
         return self.peek_every_nth(nth, foo)
 
-    def log_exceptions(self, level: str = 'error') -> Peeker:
+    def log_exceptions(self, level: str = 'error', with_trace: bool = True):
         flog = getattr(logger, level)
 
         def func(i, x):
             if is_exception(x):
-                flog('#%d:  %r', i, x)
+                trace = ''
+                if with_trace:
+                    if isinstance(x, RemoteException):
+                        trace = x.format()
+                    else:
+                        try:
+                            trace = ''.join(traceback.format_tb(x.__traceback__))
+                        except AttributeError:
+                            pass
+                if trace:
+                    flog('#%d:  %r\n%s', i, x, trace)
+                else:
+                    flog('#%d:  %r', i, x)
 
         return self.peek(func)
 
+    def batch(self, batch_size: int):
+        '''Bundle elements into batches, i.e. lists.
 
-class Stream(collections.abc.Iterator, StreamMixin):
-    @classmethod
-    def register(cls, class_: Type[Stream], name: str):
-        def f(self, *args, **kwargs):
-            return class_(self, *args, **kwargs)
-
-        setattr(cls, name, f)
-
-    def __init__(self, instream: Union[Stream, Iterator, Iterable], /):
-        if isinstance(instream, Stream):
-            self._crashed = EventUpstreamer(instream._crashed)
-            # `self._crashed` is set when the current object encounters
-            # a problem; it also signals upstream that downstream has crashed.
-            self._instream = instream
-        else:
-            self._crashed = EventUpstreamer()
-            if hasattr(instream, '__next__'):
-                self._instream = instream
-            else:
-                self._instream = iter(instream)
-        self.index = 0
-        # Index of the upcoming element; 0 based.
-        # This is also the count of finished elements.
-
-    def __iter__(self):
-        return self
-
-    def _get_next(self):
-        # Subclasses refine this method rather than `__next__`.
-        return next(self._instream)
-
-    def __next__(self):
-        z = self._get_next()
-        self.index += 1
-        return z
-
-    def collect(self) -> list:
-        return list(self)
-
-    def drain(self) -> Union[int, Tuple[int, int]]:
-        '''Drain off the stream.
-
-        Return the number of elements processed.
-        When there are exceptions, return the total number of elements
-        as well as the number of exceptions.
-        '''
-        n = 0
-        nexc = 0
-        for v in self:
-            n += 1
-            if is_exception(v):
-                nexc += 1
-        if nexc:
-            return n, nexc
-        return n
-
-    def batch(self, batch_size: int) -> Batcher:
-        '''Take elements from an input stream,
+        Take elements from an input stream,
         and bundle them up into batches up to a size limit,
         and produce the batches in an iterable.
 
-        The output batches are all of the specified size, except possibly the final batch.
-        There is no 'timeout' logic to produce a smaller batch.
+        The output batches are all of the specified size,
+        except possibly the final batch.
+        There is no 'timeout' logic to proceed eagerly with a partial batch .
         For efficiency, this requires the input stream to have a steady supply.
         If that is a concern, having a `buffer` on the input stream may help.
         '''
-        return Batcher(self, batch_size)
+        self.streamlets.append(Batcher(self.streamlets[-1], batch_size))
+        return self
 
-    def unbatch(self) -> Unbatcher:
-        '''Reverse of "batch", turning a stream of batches into
-        a stream of individual elements.
+    def unbatch(self):
+        '''Reverse of "batch".
+
+        Turn a stream of lists into a stream of individual elements.
+
+        This is usually used to correpond with a previous
+        `.batch()`, but that is not required. The only requirement
+        is that the input elements are lists.
         '''
-        return Unbatcher(self)
+        self.streamlets.append(Unbatcher(self.streamlets[-1]))
+        return self
 
-    def drop_if(self, func: Callable[[int, T], bool]) -> Dropper:
-        return Dropper(self, func)
-
-    def head(self, n: int) -> Header:
-        return Header(self, n)
-
-    def peek(self, func: Callable[[int, T], None] = None) -> Peeker:
-        '''Take a peek at the data element *before* it is sent
-        on for processing.
-
-        The function `func` takes the data index (0-based)
-        and the data element. Typical actions include print out
-        info or save the data for later inspection. Usually this
-        function should not modify the data element in the stream.
-        '''
-        if func is None:
-            func = _default_peek_func
-        return Peeker(self, func)
-
-    def buffer(self, maxsize: int = None) -> Buffer:
+    def buffer(self, maxsize: int = None):
         '''Buffer is used to stabilize and improve the speed of data flow.
 
         A buffer is useful after any operation that can not guarantee
@@ -355,27 +422,26 @@ class Stream(collections.abc.Iterator, StreamMixin):
         data. The buffer evens out unstabilities in the speeds of upstream
         production and downstream consumption.
         '''
-        if maxsize is None:
-            maxsize = 1024
-        else:
-            assert 1 <= maxsize <= 10_000
-        return Buffer(self, maxsize)
+        if not self.streamlets[0]._started.is_set():
+            raise RuntimeError("`buffer` requires the object to be in context manager")
+        self.streamlets.append(Buffer(self.streamlets[-1], maxsize))
+        return self
 
     def transform(self,
-                  func: Callable[[T], TT],
+                  func: Union[Callable[[T], TT], Callable[[T], Awaitable[TT]]],
                   *,
-                  workers: Optional[Union[int, str]] = None,
+                  concurrency: Optional[Union[int, str]] = None,
+                  return_x: bool = False,
                   return_exceptions: bool = False,
-                  keep_order: bool = None,
-                  **func_args) -> Union[Transformer, ConcurrentTransformer]:
+                  **func_args):
         '''Apply a transformation on each element of the data stream,
         producing a stream of corresponding results.
 
-        `func`: a sync function that takes a single input item
+        `func`: a sync or async function that takes a single input item
         as the first positional argument and produces a result.
         Additional keyword args can be passed in via `func_args`.
 
-        The outputs are in the order of the input elements in `self._instream`.
+        The outputs are in the order of the input elements.
 
         The main point of `func` does not have to be the output.
         It could rather be some side effect. For example,
@@ -386,35 +452,77 @@ class Stream(collections.abc.Iterator, StreamMixin):
         Exception objects (if `return_exceptions` is `True`), which may be
         counted, logged, or handled in other ways.
 
-        `workers`: max number of concurrent calls to `func`. By default
-        this is 0, i.e. there is no concurrency.
+        `concurrency`: max number of concurrent calls to `func`. By default
+        there is no concurrency, but this is usually *not* what you want,
+        because the point of this method is to run an I/O-bound operation
+        with concurrency.
 
-        `workers=0` and `workers=1` are different. The latter runs the
-        transformer in a separate thread whereas the former runs "inline".
+        `return_x`: if True, output stream will contain tuples `(x, y)`;
+        if False, output stream will contain `y` only.
 
-        When `workers = N > 0`, the worker threads are named 'transformer-0',
-        'transformer-1',..., 'transformer-<N-1>'.
+        `return_exceptions`: if True, exceptions raised by `func` will be
+        in the output stream as if they were regular results; if False,
+        they will propagate. Note that this does not absorbe exceptions
+        raised by previous components in the pipeline; it is concered about
+        exceptions raised by `func` only.
 
-        `keep_order`: whether the output elements should be ordered the same
-            as the corresponding input elements.
-            This is ignored when `workers` is `None` or 0 or 1, because
-            in those cases order is always preserved.
+        User may want to add a `buffer` to the output of this method,
+        esp if the `func` operations are slow.
         '''
-        if func_args:
-            func = functools.partial(func, **func_args)
-
-        if workers is None or workers == 0:
-            return Transformer(self, func, return_exceptions=return_exceptions)
-
-        if workers == 'max':
-            workers = MAX_THREADS
-        else:
-            1 <= workers <= 100
-        return ConcurrentTransformer(
-            self, func, workers=workers,
+        if not self.streamlets[0]._started.is_set():
+            raise RuntimeError("`transform` requires the object to be in context manager")
+        self.streamlets.append(Transformer(
+            self.streamlets[-1],
+            func,
+            concurrency=concurrency,
+            return_x=return_x,
             return_exceptions=return_exceptions,
-            keep_order=keep_order,
-        )
+            **func_args,
+        ))
+        return self
+
+
+class Stream(EnforceOverrides):
+    def __init__(self, instream: Union[Stream, Iterator, Iterable], /):
+        if isinstance(instream, Stream):
+            self._instream = instream
+            self._thread_pool = instream._thread_pool
+            self._started = instream._started
+            self._stopped = instream._stopped
+        else:
+            if hasattr(instream, '__next__'):
+                self._instream = instream
+            else:
+                self._instream = iter(instream)
+            self._thread_pool: concurrent.futures.ThreadPoolExecutor = None
+            self._started = threading.Event()
+            self._stopped = threading.Event()
+        self.index = 0
+        # Index of the upcoming element; 0 based.
+        # This is also the count of finished elements.
+
+    @final
+    def __iter__(self):
+        if not self._started.is_set():
+            raise RuntimeError("iteration must be started within context manager, i.e. within a 'with ...:' block")
+        return self
+
+    def _get_next(self):
+        '''Produce the next element in the stream.
+
+        Subclasses refine this method rather than `__next__`.
+        In a subclass, almost always it will not get the next element
+        from `_instream`, but rather from some object that holds
+        results of transformations on its `_instream`.
+        Subclass should take efforts to handle exceptions in this method.
+        '''
+        return next(self._instream)
+
+    @final
+    def __next__(self):
+        z = self._get_next()
+        self.index += 1
+        return z
 
 
 class Batcher(Stream):
@@ -474,7 +582,7 @@ class Dropper(Stream):
 class Header(Stream):
     def __init__(self, instream: Stream, n: int):
         super().__init__(instream)
-        assert n >= 0
+        assert n > 0
         self.n = n
 
     @overrides
@@ -496,53 +604,76 @@ class Peeker(Stream):
         return z
 
 
+GET_SLEEP = 0.00026
+PUT_SLEEP = 0.00015
+
+
+def put_in_queue(q, x, stop_event):
+    while True:
+        try:
+            q.put(x, timeout=PUT_SLEEP)
+            return True
+        except QueueFull:
+            if stop_event.is_set():
+                return False
+
+
+async def a_put_in_queue(q, x, stop_event):
+    while True:
+        try:
+            q.put_nowait(x)
+            return True
+        except QueueFull:
+            if stop_event.is_set():
+                return False
+            await asyncio.sleep(PUT_SLEEP)
+
+
+def get_from_queue(q, stop_event, worker):
+    while True:
+        try:
+            return q.get(timeout=GET_SLEEP)
+        except QueueEmpty:
+            if worker.done():
+                if worker.exception():
+                    raise worker.exception()
+                assert stop_event.is_set()
+                return
+
+
 class Buffer(Stream):
-    def __init__(self, instream: Stream, maxsize: int):
+    def __init__(self, instream: Stream, maxsize: int = None):
         super().__init__(instream)
-        assert 1 <= maxsize <= 10_000
+        if maxsize is None:
+            maxsize = 1024
+        else:
+            assert 1 <= maxsize <= 10_000
         self.maxsize = maxsize
-        self._q = IterQueue(maxsize, downstream_crashed=self._crashed)
-        self._upstream_err = None
-        self._thread = None
-        self._start()
+        self._q = Queue(maxsize)
+        self._nomore = object()
+        self._t = self._thread_pool.submit(self._start)
+        # This requires that `instream` is already context managed.
 
     def _start(self):
-        def foo(instream, q, crashed):
-            try:
-                for v in instream:
-                    if crashed.is_set():
-                        break
-                    q.put(v)
-                q.close()
-            except Exception as e:
-                # This should be exception while
-                # getting data from `instream`,
-                # not exception in the current object.
-                self._upstream_err = e
-                q.close()
+        threading.current_thread().name = 'BufferThread'
+        q = self._q
+        instream = self._instream
+        stopped = self._stopped
 
-        self._thread = threading.Thread(
-            target=foo, args=(self._instream, self._q, self._crashed),
-            name='BufferThread')
-        self._thread.start()
-
-    def _stop(self):
-        if self._thread is not None:
-            self._thread.join()
-            self._thread = None
-
-    def __del__(self):
-        self._stop()
+        for v in instream:
+            # If error happens in the line above,
+            # it will crash this thread and be reflected in the
+            # concurrent.futures.Future object.
+            if not put_in_queue(q, v, stopped):
+                return
+        put_in_queue(q, self._nomore, stopped)
 
     @overrides
     def _get_next(self):
-        try:
-            z = next(self._q)
-            return z
-        except StopIteration:
-            if self._upstream_err is not None:
-                raise self._upstream_err
-            raise
+        z = get_from_queue(self._q, self._stopped, self._t)
+        if z is self._nomore:
+            raise StopIteration
+        return z
 
 
 class Transformer(Stream):
@@ -550,169 +681,84 @@ class Transformer(Stream):
                  instream: Stream,
                  func: Callable[[T], TT],
                  *,
+                 concurrency: Union[int, str] = None,
+                 return_x: bool = False,
                  return_exceptions: bool = False,
+                 **func_args,
                  ):
         super().__init__(instream)
-        self.func = func
-        self.return_exceptions = return_exceptions
+
+        if concurrency is None:
+            concurrency = 1
+        elif concurrency == 'max':
+            concurrency = MAX_THREADS
+        else:
+            assert concurrency > 0
+        self._return_x = return_x
+        self._return_exceptions = return_exceptions
+        self._nomore = object()
+        self._tasks = Queue(concurrency)
+
+        if is_async(func):
+            self._t = self._thread_pool.submit(self._start_async, func, **func_args)
+        else:
+            self._t = self._thread_pool.submit(self._start_sync, func, **func_args)
+
+    def _start_sync(self, func, **kwargs):
+        tasks = self._tasks
+        pool = self._thread_pool
+        stopped = self._stopped
+        for x in self._instream:
+            t = pool.submit(func, x, **kwargs)
+            if not put_in_queue(tasks, (x, t), stopped):
+                return
+        put_in_queue(tasks, self._nomore, stopped)
+
+    def _start_async(self, func, **kwargs):
+        async def main():
+            tasks = self._tasks
+            ff = func
+            args = kwargs
+            stopped = self._stopped
+
+            for x in self._instream:
+                t = asyncio.create_task(ff(x, **args))
+                if not (await a_put_in_queue(tasks, (x, t), stopped)):
+                    return
+            if not (await a_put_in_queue(tasks, self._nomore, stopped)):
+                return
+
+            # If do not wait, `main` will exit, and unfinished tasks
+            # will be cancelled.
+            while not tasks.empty():
+                if stopped.is_set():
+                    return
+                await asyncio.sleep(0.001)
+
+        asyncio.run(main())
 
     @overrides
     def _get_next(self):
-        z = next(self._instream)
-        try:
-            return self.func(z)
-        except Exception as e:
-            if self.return_exceptions:
+        z = get_from_queue(self._tasks, self._stopped, self._t)
+        if z is self._nomore:
+            raise StopIteration
+        if z is None:
+            return
+
+        x, fut = z
+        while not fut.done():
+            if self._stopped.is_set():
+                return
+            sleep(GET_SLEEP)
+        if fut.exception():
+            e = fut.exception()
+            if self._return_exceptions:
+                if self._return_x:
+                    return x, e
                 return e
-            self._crashed.set()
-            raise
-
-
-def transform(
-        in_stream: Iterator, out_stream: IterQueue, func, *,
-        workers, return_exceptions, keep_order, crashed,
-        upstream_err, finished=None):
-
-    def _process(in_stream, out_stream, func, lock, finished, crashed, keep_order):
-        # `finished`: finished by one of these workers, either due to
-        # exhaustion of input, or error in processing.
-        # `crashed`: some downstream transformer failed, while the current
-        # may be all fine.
-
-        def set_finish():
-            active_tasks.pop()
-            if not active_tasks:
-                # This thread is the last active one
-                # for the current transformer.
-                out_stream.close()
-            finished.set()
-
-        Future = concurrent.futures.Future
-        while True:
-            with lock:
-                if finished.is_set():
-                    set_finish()
-                    return
-                if crashed.is_set():
-                    set_finish()
-                    return
-
-                try:
-                    x = next(in_stream)
-                except StopIteration:
-                    set_finish()
-                    return
-                except Exception as e:
-                    upstream_err.append(e)
-                    set_finish()
-                    return
-                else:
-                    if keep_order:
-                        # The lock ensures the output future
-                        # holds the spot for the input
-                        # in the correct order.
-                        fut = Future()
-                        out_stream.put(fut)
-
-            if keep_order:
-                try:
-                    y = func(x)
-                except Exception as e:
-                    if return_exceptions:
-                        fut.set_result(e)
-                    else:
-                        fut.set_exception(e)
-                        crashed.set()
-                        set_finish()
-                        return
-                else:
-                    fut.set_result(y)
-            else:
-                try:
-                    y = func(x)
-                except Exception as e:
-                    out_stream.put(e)
-                    if not return_exceptions:
-                        crashed.set()
-                        set_finish()
-                        return
-                else:
-                    out_stream.put(y)
-
-    if finished is None:
-        finished = threading.Event()
-    lock = threading.Lock()
-    tasks = [
-        threading.Thread(
-            target=_process,
-            name=f'TransformerThread-{i}',
-            args=(in_stream, out_stream, func, lock, finished, crashed, keep_order),
-        )
-        for i in range(workers)
-    ]
-    active_tasks = list(range(workers))
-
-    for t in tasks:
-        t.start()
-    return tasks
-
-
-class ConcurrentTransformer(Stream):
-    def __init__(self,
-                 instream: Stream,
-                 func: Callable[[T], TT],
-                 *,
-                 workers: int,
-                 return_exceptions: bool = False,
-                 keep_order: bool = None,  # default True
-                 ):
-        assert workers >= 1
-        super().__init__(instream)
-        self.func = func
-        self.workers = workers
-        self.return_exceptions = return_exceptions
-        self.keep_order = workers > 1 and (keep_order is None or keep_order)
-        self._outstream = IterQueue(1024, downstream_crashed=self._crashed)
-        self._tasks = []
-        self._upstream_err = []
-        self._start()
-
-    def _start(self):
-        self._tasks = transform(
-            self._instream,
-            self._outstream,
-            self.func,
-            workers=self.workers,
-            return_exceptions=self.return_exceptions,
-            keep_order=self.keep_order,
-            crashed=self._crashed,
-            upstream_err=self._upstream_err,
-        )
-
-    def _stop(self):
-        for t in self._tasks:
-            t.join()
-
-    def __del__(self):
-        self._stop()
-
-    @overrides
-    def _get_next(self):
-        if self._upstream_err:
-            raise self._upstream_err[0]
-        try:
-            y = next(self._outstream)
-        except StopIteration:
-            if self._upstream_err:
-                raise self._upstream_err[0]
-            raise
-        if self.keep_order:
-            try:
-                return y.result()
-            except Exception:
-                self._crashed.set()
-                raise
-        if isinstance(y, Exception) and not self.return_exceptions:
-            self._crashed.set()
-            raise y
+            logger.error("exception '%r' happened for input '%s'", e, x)
+            raise e
+        y = fut.result()
+        if self._return_x:
+            return x, y
         return y
