@@ -63,16 +63,17 @@ import threading
 import time
 import warnings
 from abc import ABCMeta, abstractmethod
+from concurrent.futures import Future
 from multiprocessing import synchronize, Queue
+from time import perf_counter
 from typing import List, Type, Sequence, Union, Callable
-from uuid import uuid4
 
 import psutil
 from overrides import EnforceOverrides, overrides
 
-from .remote_exception import RemoteException
-from .socket import SocketServer, write_record
-from .util import forward_logs, logger_thread, IterQueue, FutureIterQueue, is_exception
+from .remote_exception import RemoteException, exit_err_msg
+from .util import forward_logs, logger_thread
+from ._streamer import put_in_queue, get_from_queue
 
 
 # Set level for logs produced by the standard `multiprocessing` module.
@@ -86,11 +87,19 @@ class TimeoutError(Exception):
 
 
 class EnqueueTimeout(TimeoutError):
-    pass
+    def __init__(self, duration, queue_idx):
+        super().__init__(duration, queue_idx)
+
+    def __str__(self):
+        return f"enqueue timed out after {self.args[0]} seconds at queue {self.args[1]}"
 
 
 class TotalTimeout(TimeoutError):
-    pass
+    def __init__(self, duration):
+        super().__init__(duration)
+
+    def __str__(self):
+        return f"total timed out after {self.args[0]} seconds"
 
 
 class Servlet(metaclass=ABCMeta):
@@ -179,7 +188,7 @@ class Servlet(metaclass=ABCMeta):
 
         self.batch_size = batch_size
         self.batch_wait_time = batch_wait_time
-        self.name = f'{self.__class__.__name__}--{multiprocessing.current_process().name}'
+        self.name = multiprocessing.current_process().name
 
     def start(self, *, q_in, q_out, q_err, q_in_lock, should_stop):
         q_err.put(self.name)
@@ -216,7 +225,6 @@ class Servlet(metaclass=ABCMeta):
     def _start_batch(self, *, q_in, q_out, q_err, q_in_lock, should_stop):
         batch_size = self.batch_size
         batch_wait_time = self.batch_wait_time
-        perf_counter = time.perf_counter
 
         batch_size_max = -1
         batch_size_min = 1000000
@@ -345,6 +353,7 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
     def __init__(self, *,
                  max_queue_size: int = None,
                  cpus: Sequence[int] = None,
+                 sequential_start: bool = True,
                  ):
         '''
         `cpus`: specifies the cpu for the "main process",
@@ -357,6 +366,7 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
         self._servlets: List[multiprocessing.Process] = []
         self._uid_to_futures = {}
 
+        self._sequential_start = sequential_start
         self._should_stop = self.MP_CLASS.Event()
         # `_add_servlet` can only be called after this.
         # In other words, subclasses should call `super().__init__`
@@ -365,28 +375,15 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
 
         self._thread_pool = concurrent.futures.ThreadPoolExecutor(128)
         self._thread_tasks = {}
-        self._async_tasks = {}
 
         self.started = 0
         if cpus:
             # Pin this coordinating thread to the specified CPUs.
             psutil.Process().cpu_affinity(cpus=cpus)
 
-    def __del__(self):
-        if self.started:
-            self.stop()
-
     def __enter__(self):
         # After adding servlets, all other methods of this object
         # should be used with context manager `__enter__`/`__exit__`
-        # or `start`/`stop`.
-        self.start()
-        return self
-
-    def __exit__(self, exc_type, exc_value, exc_traceback):
-        self.stop()
-
-    def start(self, sequential: bool = True):
         assert self._servlets
         self.started += 1
         if self.started > 1:
@@ -394,9 +391,10 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
             return
 
         t = self._thread_pool.submit(logger_thread, self._q_log)
+        t.add_done_callback(self._thread_task_done_callback)
         self._thread_tasks[t] = None
 
-        if sequential:
+        if self._sequential_start:
             # Start the servlets one by one.
             n = len(self._servlets)
             k = 0
@@ -424,11 +422,19 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
                 logger.info(f"{k}/{n}: servlet <{name}> is ready")
 
         t = self._thread_pool.submit(self._gather_output)
+        t.add_done_callback(self._thread_task_done_callback)
         self._thread_tasks[t] = None
         t = self._thread_pool.submit(self._gather_error)
+        t.add_done_callback(self._thread_task_done_callback)
         self._thread_tasks[t] = None
 
-    def stop(self):
+        return self
+
+    def __exit__(self, exc_type=None, exc_value=None, exc_traceback=None):
+        msg = exit_err_msg(self, exc_type, exc_value, exc_traceback)
+        if msg:
+            logger.error(msg)
+
         assert self.started > 0
         self.started -= 1
         if self.started > 0:
@@ -446,10 +452,6 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
         self._thread_pool.shutdown()
         # TODO: in Python 3.9+, use argument `cancel_futures=True`.
 
-        for t in self._async_tasks:
-            if not t.done():
-                t.cancel()
-
         # Reset CPU affinity.
         psutil.Process().cpu_affinity(cpus=[])
 
@@ -460,12 +462,8 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
         enqueue_timeout, total_timeout = self._resolve_timeout(
             enqueue_timeout=enqueue_timeout, total_timeout=total_timeout,
         )
-        time0 = time.perf_counter()
-        uid, fut = await self._async_call_enqueue(
-            x, t0=time0, enqueue_timeout=enqueue_timeout)
-        return await self._async_call_wait_for_result(
-            uid=uid, fut=fut,
-            t0=time0, total_timeout=total_timeout)
+        uid, fut = await self._async_call_enqueue(x, enqueue_timeout=enqueue_timeout)
+        return await self._async_call_wait_for_result(uid, fut, total_timeout=total_timeout)
 
     def call(self, x, /, *,
              enqueue_timeout: Union[int, float] = None,
@@ -482,77 +480,91 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
             enqueue_timeout=enqueue_timeout, total_timeout=total_timeout,
         )
 
-        time0 = time.perf_counter()
-        uid, fut = self._call_enqueue(
-            x, t0=time0, enqueue_timeout=enqueue_timeout)
-        return self._call_wait_for_result(
-            uid=uid, fut=fut, t0=time0, total_timeout=total_timeout)
+        uid, fut = self._call_enqueue(x, enqueue_timeout=enqueue_timeout)
+        return self._call_wait_for_result(uid, fut, total_timeout=total_timeout)
 
     def stream(self, data_stream, /, *,
                return_exceptions: bool = False,
                return_x: bool = False,
-               enqueue_timeout: Union[int, float] = None,
-               total_timeout: Union[int, float] = None,
                backlog: int = 1024,
+               enqueue_timeout=None,
+               total_timeout=None,
                ):
         # The order of elements in the stream is preserved, i.e.,
         # elements in the output stream corresponds to elements
         # in the input stream in the same order.
 
-        enqueue_timeout, total_timeout = self._resolve_timeout(
-            enqueue_timeout=enqueue_timeout,
-            total_timeout=total_timeout,
-        )
+        # For streaming, "timeout" is usually not a concern.
+        # The concern is overall throughput.
 
-        data_in_pipe = IterQueue(backlog)
-        results = FutureIterQueue(
-            backlog, return_x=return_x, return_exceptions=return_exceptions)
+        enqueue_timeout, total_timeout = self._resolve_timeout(
+            enqueue_timeout=enqueue_timeout, total_timeout=total_timeout)
+        tasks = queue.Queue(backlog)
+        nomore = object()
 
         def _enqueue():
-            Future = concurrent.futures.Future
             enqueue = self._call_enqueue
-            perf_counter = time.perf_counter
+            tt = tasks
+            ret_ex = return_exceptions
             et = enqueue_timeout
+            should_stop = self._should_stop
             for x in data_stream:
-                time0 = perf_counter()
                 try:
-                    uid, fut = enqueue(x, t0=time0, enqueue_timeout=et)
+                    uid, fut = enqueue(x, enqueue_timeout=et)
                 except Exception as e:
-                    uid, fut = None, e
-                ff = Future()
-                data_in_pipe.put((x, time0, uid, fut, ff))
-                results.put(ff)
-            data_in_pipe.close()
-            results.close()
-
-        def _wait():
-            tt = total_timeout
-            wait_for_result = self._call_wait_for_result
-            for x, t0, uid, fut, ff in data_in_pipe:
-                if is_exception(fut):
-                    ff.set_result((x, fut))
-                else:
-                    try:
-                        y = wait_for_result(uid=uid, fut=fut, t0=t0, total_timeout=tt)
-                    except Exception as e:
-                        y = e
-                    ff.set_result((x, y))
-
-        def cb(fut):
-            try:
-                del self._thread_tasks[fut]
-            except KeyError:
-                warnings.warn(f"the Future object `{fut}` was already removed")
+                    if ret_ex:
+                        uid = None
+                        fut = Future()
+                        fut.set_exception(e)
+                    else:
+                        logger.error("exception '%r' happened for input '%s'", e, x)
+                        raise
+                if not put_in_queue(tt, (x, uid, fut), should_stop):
+                    return
+            put_in_queue(tt, nomore, should_stop)
 
         t = self._thread_pool.submit(_enqueue)
-        t.add_done_callback(cb)
+        t.add_done_callback(self._thread_task_done_callback)
         self._thread_tasks[t] = None
 
-        t = self._thread_pool.submit(_wait)
-        t.add_done_callback(cb)
-        self._thread_tasks[t] = None
+        _wait = self._call_wait_for_result
 
-        return results
+        while True:
+            z = get_from_queue(tasks, self._should_stop, t)
+            if z is nomore:
+                break
+            if z is None:
+                return
+
+            x, uid, fut = z
+            if uid is None:
+                if return_x:
+                    yield x, fut.exception()
+                else:
+                    yield fut.exception()
+            else:
+                try:
+                    y = _wait(uid, fut, total_timeout=total_timeout)
+                except Exception as e:
+                    if return_exceptions:
+                        if return_x:
+                            yield x, e
+                        else:
+                            yield e
+                    else:
+                        logger.error("exception '%r' happened for input '%s'", e, x)
+                        raise
+                else:
+                    if return_x:
+                        yield x, y
+                    else:
+                        yield y
+
+    def _thread_task_done_callback(self, t):
+        try:
+            del self._thread_tasks[t]
+        except KeyError:
+            warnings.warn(f"the Future object `{t}` was already removed")
 
     def _add_servlet(self,
                      servlet: Type[Servlet],
@@ -573,7 +585,7 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
 
         cpus = self._resolve_cpus(cpus=cpus, workers=workers)
 
-        name = name or 'Servlet'
+        name = name or servlet.__name__
 
         for cpu in cpus:
             # Pinned to the specified cpu core.
@@ -657,32 +669,12 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
         '''
         if enqueue_timeout is None:
             enqueue_timeout = self.TIMEOUT_ENQUEUE
-        assert 0 <= enqueue_timeout <= 10
         if total_timeout is None:
             total_timeout = max(self.TIMEOUT_TOTAL, enqueue_timeout * 10)
-        assert 0 < total_timeout <= 100
         if enqueue_timeout > total_timeout:
             enqueue_timeout = total_timeout
         return enqueue_timeout, total_timeout
-
-    @abstractmethod
-    def _init_future_struct(self, x, fut):
-        # We store the Future object for future retrieval.
-        # Different types of services may need to store additional things
-        # along with other things.
-        # Subclasses customize this method to their need.
-        raise NotImplementedError
-
-    def _enqueue_future(self, x, fut):
-        z = self._init_future_struct(x, fut)
-        while True:
-            uid = str(uuid4())
-            if self._uid_to_futures.setdefault(uid, z) is z:
-                # Stored `uid` as value `z`.
-                # If not in this block, it's because `uid`
-                # already exists in `self._uid_to_futures`,
-                # so will generate a new one.
-                return uid
+        # Accidental negative values for these are OK.
 
     @abstractmethod
     def _input_queues(self):
@@ -701,7 +693,7 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
         should_stop = self._should_stop
         while self.started:
             try:
-                uid, err = q_err.get(timeout=1)
+                uid, err = q_err.get(timeout=0.1)
             except queue.Empty:
                 if should_stop.is_set():
                     break
@@ -713,18 +705,25 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
                     continue
                 try:
                     # `err` is a RemoteException object.
-                    fut['fut'].set_exception(err)
+                    fut.set_exception(err)
                 except asyncio.InvalidStateError as e:
-                    if fut['fut'].cancelled():
+                    if fut.cancelled():
                         logger.warning('Future object is already cancelled')
                     else:
                         logger.exception(e)
                         raise
 
-    async def _async_call_enqueue(self, x, *, t0, enqueue_timeout) -> tuple:
-        loop = asyncio.get_running_loop()
-        fut = loop.create_future()
-        uid = self._enqueue_future(x, fut)
+    def _enqueue_future(self, x):
+        t0 = perf_counter()
+        fut = Future()
+        uid = id(fut)
+        self._uid_to_futures[uid] = fut
+        fut.data = {'t0': t0}
+        return uid, fut
+
+    async def _async_call_enqueue(self, x, *, enqueue_timeout):
+        uid, fut = self._enqueue_future(x)
+        t0 = fut.data['t0']
 
         qs = self._input_queues()
         t1 = t0 + enqueue_timeout
@@ -734,42 +733,48 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
                     q.put_nowait((uid, x))
                     break
                 except queue.Full:
-                    timenow = time.perf_counter()
+                    timenow = perf_counter()
                     if timenow >= t1:
                         fut.cancel()
-                        del self._uid_to_futures[uid]
-                        raise EnqueueTimeout(
-                            f'waited {timenow - t0} seconds; timeout at queue {iq}')
+                        try:
+                            del self._uid_to_futures[uid]
+                        except KeyError:
+                            print(iq, q)
+                            raise
+                        raise EnqueueTimeout(timenow - t0, iq)
                     await asyncio.sleep(min(self.SLEEP_ENQUEUE, t1 - timenow))
 
+        fut.data['t1'] = perf_counter()
         return uid, fut
 
-    async def _async_call_wait_for_result(self, *, uid, fut, t0, total_timeout):
-        t1 = t0 + total_timeout
+    async def _async_call_wait_for_result(self, uid, fut, *, total_timeout):
+        t0 = fut.data['t0']
+        t2 = t0 + total_timeout
         while not fut.done():
-            timenow = time.perf_counter()
-            if time.perf_counter() >= t1:
+            timenow = perf_counter()
+            if timenow >= t2:
                 fut.cancel()
                 self._uid_to_futures.pop(uid, None)
                 # `uid` could have been deleted by
                 # `gather_results` during very subtle
                 # timing coincidence.
-                raise TotalTimeout(f'waited {timenow - t0} seconds')
-            await asyncio.sleep(min(self.SLEEP_DEQUEUE, t1 - timenow))
-        # await asyncio.wait_for(fut, timeout=t0 + total_timeout - time.perf_counter())
+                raise TotalTimeout(timenow - t0)
+            await asyncio.sleep(min(self.SLEEP_DEQUEUE, t2 - timenow))
+        # await asyncio.wait_for(fut, timeout=t0 + total_timeout - perf_counter())
         # NOTE: `asyncio.wait_for` seems to be blocking for the
         # `timeout` even after result is available.
+        fut.data['t2'] = perf_counter()
         return fut.result()
         # This could raise RemoteException.
 
-    def _call_enqueue(self, x, *, t0, enqueue_timeout) -> tuple:
-        fut = concurrent.futures.Future()
-        uid = self._enqueue_future(x, fut)
+    def _call_enqueue(self, x, *, enqueue_timeout):
+        uid, fut = self._enqueue_future(x)
+        t0 = fut.data['t0']
 
         qs = self._input_queues()
         t1 = t0 + enqueue_timeout
         for iq, q in enumerate(qs):
-            timenow = time.perf_counter()
+            timenow = perf_counter()
             try:
                 # `enqueue_timeout` is the combined time across input queues,
                 # not time per queue.
@@ -780,22 +785,25 @@ class MPServer(EnforceOverrides, metaclass=ABCMeta):
             except queue.Full:
                 fut.cancel()
                 del self._uid_to_futures[uid]
-                raise EnqueueTimeout(
-                    f'waited {time.perf_counter() - t0} seconds; timeout at queue {iq}')
+                raise EnqueueTimeout(perf_counter() - t0, iq)
 
+        fut.data['t1'] = perf_counter()
         return uid, fut
 
-    def _call_wait_for_result(self, *, uid, fut, t0, total_timeout):
-        t1 = t0 + total_timeout
+    def _call_wait_for_result(self, uid, fut, *, total_timeout):
+        t0 = fut.data['t0']
+        t2 = t0 + total_timeout
         try:
-            return fut.result(timeout=t1 - time.perf_counter())
+            res = fut.result(timeout=t2 - perf_counter())
         except concurrent.futures.TimeoutError:
             fut.cancel()
             self._uid_to_futures.pop(uid, None)
             # `uid` could have been deleted by
             # `gather_results` during very subtle
             # timing coincidence.
-            raise TotalTimeout(f'waited {time.perf_counter() - t0} seconds')
+            raise TotalTimeout(perf_counter() - t0)
+        fut.data['t2'] = perf_counter()
+        return res
 
 
 class SequentialServer(MPServer):
@@ -841,7 +849,7 @@ class SequentialServer(MPServer):
                     # timed-out in `async_call` or `call`.
                     continue
                 try:
-                    fut['fut'].set_result(y)
+                    fut.set_result(y)
                 except asyncio.InvalidStateError as e:
                     if fut.cancelled():
                         logger.warning('Future object is already cancelled')
@@ -849,10 +857,6 @@ class SequentialServer(MPServer):
                         logger.exception(e)
                         raise
 
-    @overrides
-    def _init_future_struct(self, x, fut):
-        return {'fut': fut}
-
     @overrides
     def _input_queues(self):
         return self._q_in_out[:1]
@@ -896,6 +900,12 @@ class EnsembleServer(MPServer):
         '''
         raise NotImplementedError
 
+    @overrides
+    def _enqueue_future(self, x):
+        uid, fut = super()._enqueue_future(x)
+        fut.data.update(x=x, res=[None] * len(self._q_out), n=0)
+        return uid, fut
+
     @overrides
     def _gather_output(self):
         threading.current_thread().name = 'ResultCollector'
@@ -913,35 +923,27 @@ class EnsembleServer(MPServer):
                         # timed-out in `async_call` or `call`.
                         continue
 
-                    fut['res'][idx] = y
-                    fut['n'] += 1
-                    if fut['n'] == n_results_needed:
+                    extra = fut.data
+                    extra['res'][idx] = y
+                    extra['n'] += 1
+                    if extra['n'] == n_results_needed:
                         # All results for this request have been collected.
                         del futures[uid]
                         try:
-                            z = self.ensemble(fut['x'], fut['res'])
-                            fut['fut'].set_result(z)
+                            z = self.ensemble(extra['x'], extra['res'])
+                            fut.set_result(z)
                         except asyncio.InvalidStateError as e:
-                            if fut['fut'].cancelled():
+                            if fut.cancelled():
                                 logger.warning(
                                     'Future object is already cancelled')
                             else:
                                 logger.exception(e)
                                 raise
                         except Exception as e:
-                            fut['fut'].set_exception(e)
+                            fut.set_exception(e)
                     # No sleep. Get results out of the queue as quickly as possible.
             time.sleep(self.SLEEP_DEQUEUE)
 
-    @overrides
-    def _init_future_struct(self, x, fut):
-        return {
-            'x': x,
-            'fut': fut,
-            'res': [None] * len(self._q_out),
-            'n': 0
-        }
-
     @overrides
     def _input_queues(self):
         return self._q_in
@@ -986,39 +988,3 @@ class SimpleServer(SequentialServer):
                          batch_size=batch_size,
                          func=func,
                          **kwargs)
-
-
-class MPSocketServer(SocketServer):
-    def __init__(self, server: MPServer, **kwargs):
-        super().__init__(**kwargs)
-        self._server = server
-        self._enqueue_timeout, self._total_timeout = server._resolve_timeout()
-
-    def __repr__(self):
-        return f'{self.__class__.__name__}({repr(self._server)})'
-
-    def __str__(self):
-        return self.__repr__()
-
-    @overrides
-    async def set_server_option(self, name: str, value):
-        if name == 'timeout':
-            self._enqueue_timeout, self._total_timeout = self._server._resolve_timeout(
-                enqueue_timeout=value[0], total_timeout=value[1])
-            return
-        await super().set_server_option(name, value)
-
-    @overrides
-    async def before_startup(self):
-        self._server.__enter__()
-
-    @overrides
-    async def after_shutdown(self):
-        self._server.__exit__(None, None, None)
-
-    @overrides
-    async def handle_request(self, data, writer):
-        y = await self._server.async_call(
-            data, enqueue_timeout=self._enqueue_timeout,
-            total_timeout=self._total_timeout)
-        await write_record(writer, y, encoder=self._encoder)
